"""Tests for advanced YAML features."""

from typing import Type

import pytest
from pydantic import BaseModel, Field

from pydantic_yaml import YamlModel

uses_refs = """
# Thanks to https://stackoverflow.com/a/2063741
bill-to: &id001
    given  : Chris
    family : Dumars
ship-to: *id001
"""


class _Name(BaseModel):
    """First/last names."""

    given: str
    family: str


class UsesRefs(YamlModel):
    """Example for the reference data."""

    bill_to: _Name = Field(alias="bill-to")
    ship_to: _Name = Field(alias="ship-to")

    class Config:
        allow_population_by_field_name = True


@pytest.mark.parametrize(("raw", "model_type"), [(uses_refs, UsesRefs),])
def test_yaml_refs(raw: str, model_type: Type[YamlModel]):
    """Tests that references are properly parsed."""
    model_type.parse_raw(raw)




from typing import Optional, Tuple, Type
import warnings

from pydantic import validator

from pydantic_yaml.model import YamlModel
from .semver import SemVer

__all__ = ["VersionedYamlModel"]


def _chk_between(v, lo=None, hi=None):
    if v is None:
        return
    if (hi is not None) and (v > hi):
        raise ValueError(f"Default version higher than maximum: {v} > {hi}")
    if (lo is not None) and (v < lo):
        raise ValueError(f"Default version lower than minimum: {v} < {lo}")


def _get_minmax_robust(
    cls: Type["VersionedYamlModel"],
) -> Tuple[Optional[SemVer], Optional[SemVer]]:
    min_, max_ = None, None
    for supcls in cls.mro():
        Config = getattr(supcls, "Config", None)
        if Config is not None:
            if min_ is None:
                min_ = getattr(Config, "min_version", None)
            if max_ is None:
                max_ = getattr(Config, "max_version", None)
    return min_, max_


class VersionedYamlModel(YamlModel):
    """YAML model with versioning support.

    Usage
    -----
    Inherit from this class, and set a Config class with attributes 
    `min_version` and/or `max_version`:

    ```python
    class MyModel(VersionedYamlModel):
        class Config:
            min_version = "1.0.0"

        foo: str = "bar"
    ```

    By default, the minimum version is "0.0.0" and the maximum is unset.
    This pattern enables you to more easily version your YAML files and
    protect against accidentally using older (or newer) configuration file formats.

    Notes
    -----
    By default, a validator checks that the "version" field is between 
    `Config.min_version` and `Config.max_version`, if those are not None.

    It's probably best not to even set the `version` field by hand, but rather
    in your configuration.
    """

    version: SemVer

    def __init_subclass__(cls) -> None:
        # Check Config class types
        Config = getattr(cls, "Config", None)
        if Config is not None:
            # check one field
            minv = getattr(Config, "min_version", None)
            if minv is not None:
                if not isinstance(minv, SemVer):
                    setattr(Config, "min_version", SemVer(minv))
            # check other field
            maxv = getattr(Config, "max_version", None)
            if maxv is not None:
                if not isinstance(maxv, SemVer):
                    setattr(Config, "max_version", SemVer(maxv))

        # Check ranges
        min_, max_ = _get_minmax_robust(cls)
        if (min_ is not None) and (max_ is not None) and (min_ > max_):
            raise ValueError(
                f"Minimum version higher than maximum: {min_!r} > {max_!r}"
            )

        # Check the default value of the "version" field
        fld = cls.__fields__["version"]
        d = fld.default
        if d is None:
            pass
        else:
            _chk_between(d, lo=min_, hi=max_)
            warnings.warn(
                f"Recommended to have `version` be required, but set default {d!r}",
                UserWarning,
            )

        if not issubclass(fld.type_, SemVer):
            raise TypeError(
                f"Field type for `version` must be SemVer, got {fld.type_!r}"
            )

    @validator("version", always=True)
    def _check_semver(cls, v):
        min_, max_ = _get_minmax_robust(cls)
        _chk_between(v, lo=min_, hi=max_)
        return v

    class Config:
        min_version = SemVer("0.0.0")
        max_version = None



"""Tests for SecretStr, SecretBytes."""

import pytest
from pydantic import SecretBytes, SecretStr

from pydantic_yaml import YamlModel


class SecretTstModel(YamlModel):
    """Normal model with secret fields."""

    ss: SecretStr
    sb: SecretBytes


def test_secret_types_no_rt():
    """Basic tests for SecretStr, SecretBytes not being dumped.
    
    Note that we actually intend to lose data when saving!
    """
    # Create model
    sm = SecretTstModel(ss="123", sb="321")
    assert sm.ss.get_secret_value() == "123"
    assert sm.sb.get_secret_value() == b"321"

    # Roundtrip (should technically work)
    y = sm.yaml()
    mdl = SecretTstModel.parse_raw(y)

    # Hey, we don't have the same values roundtrip!
    # These should be "**********", but let's not be exact - styling can change. :)
    assert mdl.ss.get_secret_value() != "123"
    assert mdl.sb.get_secret_value() != b"321"


class SecretTstModelDumpable(SecretTstModel):
    """Round-trippable model.
    
    TODO: Support YAML encoders? Or just use the JSON encoders instead?
    """

    class Config:
        json_encoders = {
            SecretStr: lambda v: v.get_secret_value() if v else None,
            SecretBytes: lambda v: v.get_secret_value() if v else None,
        }


@pytest.mark.xfail(reason="YAML/JSON encoder support doesn't exist yet.")
def test_secret_types_roundtrippable():
    """Tests roundtrippability when encoders are specified."""
    # Create newer model
    sd = SecretTstModelDumpable(ss="123", sb="321")
    y = sd.yaml()
    sd2 = SecretTstModelDumpable.parse_raw(y)
    sm_alt = SecretTstModel.parse_raw(y)

    assert sd2 == sd
    assert sm_alt.ss.get_secret_value() == sd.ss.get_secret_value()
    assert sm_alt.sb.get_secret_value() == sd.sb.get_secret_value()



from functools import wraps
from typing import Any, Callable, Optional, Union, no_type_check

from pydantic import errors
from pydantic_yaml.compat.types import YamlStr
from semver import VersionInfo

__all__ = ["SemVer"]


Comparator = Callable[["SemVer", Any], bool]


def _comparator(operator: Comparator) -> Comparator:
    """Wrap a Version binary op method in a type-check."""

    @wraps(operator)
    def wrapper(self: "SemVer", other: Any) -> bool:
        if not isinstance(other, SemVer):
            try:
                other = SemVer(other)
            except Exception:
                return NotImplemented
        return operator(self, other)

    return wrapper


class SemVer(YamlStr):  # want to inherit from VersionInfo, but metaclass conflict
    """Semantic Version string for Pydantic.

    Depends on `semver` versions 2 or 3. Look here for how to fix the version:
    https://python-semver.readthedocs.io/en/3.0.0-dev.2/install.html#release-policy
    """

    allow_build: bool = True
    allow_prerelease: bool = True

    __slots__ = ["_info"]

    @no_type_check
    def __new__(cls, version: Optional[str], **kwargs) -> object:
        return YamlStr.__new__(cls, cls.parse(**kwargs) if version is None else version)

    def __init__(self, version: str):
        str.__init__(version)
        self._info = VersionInfo.parse(version)

    @classmethod
    def parse(
        self,
        major: int,
        minor: int = 0,
        patch: int = 0,
        prerelease: Optional[str] = None,
        build: Optional[str] = None,
    ) -> str:
        return str(VersionInfo(major, minor, patch, prerelease, build))

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "SemVer"]) -> "SemVer":
        vi = VersionInfo.parse(value)
        if not cls.allow_build and (vi.build is None):
            raise errors.NotNoneError()
        if not cls.allow_prerelease and (vi.prerelease is None):
            raise errors.NotNoneError()
        return cls(value)

    @property
    def info(self) -> VersionInfo:
        return self._info

    @info.setter
    def info(self, value):
        raise AttributeError("attribute 'info' is readonly")

    @property
    def major(self) -> int:
        """The major part of a version (read-only)."""
        return self._info.major

    @major.setter
    def major(self, value):
        raise AttributeError("attribute 'major' is readonly")

    @property
    def minor(self) -> int:
        """The minor part of a version (read-only)."""
        return self._info.minor

    @minor.setter
    def minor(self, value):
        raise AttributeError("attribute 'minor' is readonly")

    @property
    def patch(self) -> int:
        """The patch part of a version (read-only)."""
        return self._info.patch

    @patch.setter
    def patch(self, value):
        raise AttributeError("attribute 'patch' is readonly")

    @property
    def prerelease(self) -> Optional[str]:
        """The prerelease part of a version (read-only)."""
        return self._info.prerelease

    @prerelease.setter
    def prerelease(self, value):
        raise AttributeError("attribute 'prerelease' is readonly")

    @property
    def build(self) -> Optional[str]:
        """The build part of a version (read-only)."""
        return self._info.build

    @build.setter
    def build(self, value):
        raise AttributeError("attribute 'build' is readonly")

    def __hash__(self) -> int:
        return super.__hash__(self)  # use string hashing

    @_comparator
    def __eq__(self, other: "SemVer"):
        return self._info == other._info

    @_comparator
    def __ne__(self, other: "SemVer"):
        return self._info != other._info

    @_comparator
    def __lt__(self, other: "SemVer"):
        return self._info < other._info

    @_comparator
    def __le__(self, other: "SemVer"):
        return self._info <= other._info

    @_comparator
    def __gt__(self, other: "SemVer"):
        return self._info > other._info

    @_comparator
    def __ge__(self, other: "SemVer"):
        return self._info >= other._info



"""YAML-enabled Pydantic models."""

__all__ = [
    "__version__",
    "yaml",
    "yaml_safe_dump",
    "yaml_safe_load",
    "YamlEnum",
    "YamlInt",
    "YamlIntEnum",
    "YamlStr",
    "YamlStrEnum",
    "YamlModel",
    "YamlModelMixin",
    "YamlModelMixinConfig",
    "SemVer",
    "VersionedYamlModel",
]

from .compat.old_enums import YamlEnum
from .compat.hacks import inject_all as _inject_yaml_hacks
from .compat.types import YamlInt, YamlIntEnum, YamlStr, YamlStrEnum
from .compat.yaml_lib import yaml, yaml_safe_dump, yaml_safe_load
from .ext.semver import SemVer
from .ext.versioned_model import VersionedYamlModel
from .mixin import YamlModelMixin, YamlModelMixinConfig
from .model import YamlModel
from .version import __version__

_inject_yaml_hacks()



from pydantic import BaseModel

from .mixin import YamlModelMixin

_pre_doc = """`pydantic.BaseModel` class with built-in YAML support.

You can alternatively inherit from this to implement your model:
`(pydantic_yaml.YamlModelMixin, pydantic.BaseModel)`

See Also
--------
pydantic-yaml: https://github.com/NowanIlfideme/pydantic-yaml
pydantic: https://pydantic-docs.helpmanual.io/
pyyaml: https://pyyaml.org/
ruamel.yaml: https://yaml.readthedocs.io/en/latest/index.html
"""


class YamlModel(YamlModelMixin, BaseModel):
    __doc__ = _pre_doc + BaseModel.__doc__



"""Tests ability to have keys 'unsorted' (in field definition order)."""


def test_sorting():
    from pydantic_yaml import YamlModel

    class B(YamlModel):
        val: str = "val"

    class X(YamlModel):
        c: str = "c"
        b: B = B()
        a: int = 1

    x = X()
    assert x.yaml(sort_keys=False) == "c: c\nb:\n  val: val\na: 1\n"
    assert x.yaml(sort_keys=True) == "a: 1\nb:\n  val: val\nc: c\n"



from __future__ import annotations

from pathlib import Path
from typing import Optional

import pytest

from pydantic_yaml import YamlModel


class A(YamlModel):
    """Recursive model."""

    a: int
    inner: Optional["A"]


A.update_forward_refs()


def test_recursive_yaml():
    """Test how we react to (unsupported) recursive YAML documents."""

    file = Path(__file__).parent / "recursive.yaml"

    with pytest.raises(ValueError):
        A.parse_file(file, content_type="application/yaml")



"""YAML-enabled Pydantic models."""

from .version import __version__

try:
    __all__ = [
        "__version__",
        "SemVer",
        "yaml",
        "VersionedYamlModel",
        "YamlEnum",  # deprecated class
        "YamlInt",
        "YamlIntEnum",
        "YamlModel",
        "YamlModelMixin",
        "YamlModelMixinConfig",
        "YamlStr",
        "YamlStrEnum",
    ]
    from .main import (
        SemVer,
        VersionedYamlModel,
        YamlEnum,
        YamlInt,
        YamlIntEnum,
        YamlModel,
        YamlModelMixin,
        YamlModelMixinConfig,
        YamlStr,
        YamlStrEnum,
        yaml,
    )
except Exception:
    pass



"""Tests for YAML compatible types."""

import pytest

from pydantic_yaml.compat.types import YamlInt, YamlIntEnum, YamlStr, YamlStrEnum
from pydantic_yaml.compat.yaml_lib import yaml_safe_dump, yaml_safe_load


class XSE(YamlStrEnum):
    a = "a"
    b = "b"


class XIE(YamlIntEnum):
    a = 1
    b = 2


class XI(YamlInt):
    def __new__(cls, val):
        return super().__new__(cls, int(val))


class XS(YamlStr):
    def __new__(cls, val):
        return super().__new__(cls, f"My value is: {val!s}")


def test_str_enum():
    """Test for YamlStrEnum class."""

    x1 = XSE.a
    x2 = XSE("b")
    assert yaml_safe_load(yaml_safe_dump(x1)) == "a"
    assert yaml_safe_load(yaml_safe_dump(x2)) == "b"

    with pytest.raises(ValueError):
        XSE("c")


def test_int_enum():
    """Test for YamlIntEnum class."""
    x1 = XIE.a
    x2 = XIE(2)
    assert yaml_safe_load(yaml_safe_dump(x1)) == 1
    assert yaml_safe_load(yaml_safe_dump(x2)) == 2

    with pytest.raises(ValueError):
        XIE(3)


def test_int():
    """Test for YamlInt class."""
    x = XI("123")
    assert isinstance(x, XI)
    assert x == 123
    assert repr(x) == "XI(123)"


def test_str():
    """Test for YamlStr class."""
    x = XS("123")
    _exp = "My value is: 123"
    assert isinstance(x, XS)
    assert x == _exp
    assert repr(x) == "XS('My value is: 123')"



"""Imports an installed YAML library."""

# flake8: noqa

from typing import Any, Optional
from io import BytesIO, StringIO, IOBase


try:
    import ruamel.yaml as yaml  # type: ignore

    # ruamel.yaml doesn't have type annotations

    if yaml.__version__ < "0.15.0":
        __yaml_lib__ = "ruamel-old"
    else:
        __yaml_lib__ = "ruamel-new"
except ImportError:
    try:
        import yaml  # type: ignore

        __yaml_lib__ = "pyyaml"
    except ImportError:
        raise ImportError(
            "Could not import ruamel.yaml or pyyaml, "
            "please install at least one of them."
        )

dumper_classes = []
for _fld in dir(yaml):
    try:
        if "Dumper" in _fld:
            _obj = getattr(yaml, _fld)
            dumper_classes.append(_obj)
    except Exception:
        pass

representer_classes = []
for _fld in dir(yaml):
    try:
        if "Representer" in _fld:
            _obj = getattr(yaml, _fld)
            representer_classes.append(_obj)
    except Exception:
        pass


def yaml_safe_load(stream) -> Any:
    """Wrapper around YAML library loader."""
    if __yaml_lib__ in ["ruamel-old", "pyyaml"]:
        return yaml.safe_load(stream)
    # Fixing deprecation warning in new ruamel.yaml versions
    assert __yaml_lib__ == "ruamel-new"
    ruamel_obj = yaml.YAML(typ="safe", pure=True)
    if isinstance(stream, str):
        return ruamel_obj.load(StringIO(stream))
    elif isinstance(stream, bytes):
        return ruamel_obj.load(BytesIO(stream))
    # we hope it's a stream, but don't enforce it
    return ruamel_obj.load(stream)


def yaml_safe_dump(
    data: Any, stream=None, *, sort_keys: bool = False, **kwds
) -> Optional[Any]:
    """Wrapper around YAML library dumper.
    
    Parameters
    ----------
    data
        The data you want to dump, typically a mapping (dict).
    stream
        The stream to dump to. By default (if no stream is given, i.e. None),
        this will instead dump to a text stream.
    sort_keys : bool
        Whether to sort the keys of mappings before dumping.
        Default value is False, rather than True, which is the YAML default.
        This is because Pydantic configs are easier to read if dumped in the
        same order as they are defined.
    kwds
        Other keyword arguments to set on the YAML dumper instance.
    """
    if __yaml_lib__ in ["ruamel-old", "pyyaml"]:
        return yaml.safe_dump(data, stream=stream, sort_keys=sort_keys, **kwds)
    # Fixing deprecation warning in new ruamel.yaml versions
    assert __yaml_lib__ == "ruamel-new"
    ruamel_obj = yaml.YAML(typ="safe", pure=True)
    ruamel_obj.sort_base_mapping_type_on_output = sort_keys  # type: ignore
    # Hacking some options that aren't available
    for kw in ["encoding", "default_flow_style", "default_style", "indent"]:
        if kw in kwds:
            setattr(ruamel_obj, kw, kwds[kw])

    if stream is None:
        text_stream = StringIO()
        ruamel_obj.dump(data, stream=text_stream)
        text_stream.seek(0)  # otherwise we always get ''
        return text_stream.read()
    else:
        ruamel_obj.dump(data, stream=stream)
        return None


__all__ = ["yaml_safe_dump", "yaml_safe_load", "yaml", "__yaml_lib__"]



from pathlib import Path

import pytest
from pydantic import ValidationError

from pydantic_yaml.ext.semver import SemVer
from pydantic_yaml.ext.versioned_model import VersionedYamlModel


def test_versioned_yaml():
    """Test VersionedYamlModel."""

    file = Path(__file__).parent / "versioned.yaml"

    class A(VersionedYamlModel):
        foo: str = "bar"

    class B(VersionedYamlModel):
        foo: str = "bar"

        class Config:
            min_version = "2.0.0"

    class C(VersionedYamlModel):
        foo: str = "baz"

        class Config:
            max_version = "0.5.0"

    A.parse_file(file)

    with pytest.raises(ValidationError):
        B.parse_file(file)

    with pytest.raises(ValidationError):
        C.parse_file(file)

    with pytest.raises(ValueError):

        class BadVersionConstraint(VersionedYamlModel):
            class Config:
                min_version = "3.0.0"
                max_version = "2.1.0"


def test_versioned_docs():
    """Test docs for versioned model."""

    class A(VersionedYamlModel):
        """Model with min, max constraints as None."""

        foo: str = "bar"

    class B(VersionedYamlModel):
        """Model with a maximum version set."""

        foo: str = "bar"

        class Config:
            min_version = "2.0.0"

    ex_yml = """
    version: 1.0.0
    foo: baz
    """

    a = A.parse_raw(ex_yml)
    assert a.version == SemVer("1.0.0")
    assert a.foo == "baz"

    with pytest.raises(ValidationError):
        B.parse_raw(ex_yml)
